

#include <stdio.h>
#include <assert.h>
#include "newthreads.h"

#ifndef GNUCC
#define inline /**/
#endif

typedef struct BufferDescriptor {
  void* buf;
  int   size;
  struct BufferDescriptor* next;
} BufferDescriptor;


typedef struct {
  Thread*  waiters;
  Spinlock lock;

  BufferDescriptor *firstBuf;   /* waiting buffers */
  BufferDescriptor *lastBuf;
} CommPort;


int MyNodeNum;
int NumNodes;

#ifdef PVM
static char componentName[128];
extern int saveArgc;
extern char** saveArgv;
#endif
#ifdef PVM3
static	int	*the_nodelist;
#include <pvm3.h>
#endif

static Spinlock  thePortsLock;
static CommPort** thePorts = (CommPort**)0;
static int numPorts = 0;

int numRecvWaiters = 0;
static Spinlock numRecvWaitersLock;


static BufferDescriptor* freeBufDescs = (BufferDescriptor*)0;

extern void* malloc(int);


inline static BufferDescriptor* newBufDesc()
{
  BufferDescriptor* ret;

  if(freeBufDescs){
	ret = freeBufDescs;
	freeBufDescs = ret->next;
  }else{
	ret = (BufferDescriptor*)malloc(sizeof(BufferDescriptor));
	assert(ret);
  }
  return ret;
}


inline static void freeBufDesc(BufferDescriptor* bd)
{
  bd->next = freeBufDescs;
  freeBufDescs = bd;
}


void initCommStuff()
{
  int i;
#ifdef PVM3
  int me;
  int my_parent;
  int retval;
#endif

  InitSpinlock(thePortsLock); 
  InitSpinlock(numRecvWaitersLock);
  numPorts = 32;
  thePorts = (CommPort**)malloc(numPorts*sizeof(CommPort*));
  assert(thePorts);
  for(i=0; i<numPorts; ++i){
	thePorts[i] = (CommPort*)malloc(sizeof(CommPort));
	assert(thePorts[i]);
	InitSpinlock(thePorts[i]->lock);
	thePorts[i]->waiters = (Thread*)0;
	thePorts[i]->firstBuf = thePorts[i]->lastBuf = (BufferDescriptor*)0;
  }
#ifdef OneNode
  MyNodeNum = 0;
  NumNodes = 1;
#endif
#ifdef PVM
  whoami(componentName, (int*)0);
  rcv(0xfeed);
  getnint(&MyNodeNum, 1);
  getnint(&NumNodes, 1);
  getnint(&saveArgc, 1);
  saveArgv = (char**)malloc((saveArgc+1)*sizeof(char*));
  assert(saveArgv);
  for(i=0; i<saveArgc; ++i){
	saveArgv[i] = (char *)malloc(1024);
	getstring(saveArgv[i], 1);
  }
  saveArgv[saveArgc] = (char*)0;
#endif
#ifdef PVM3
  me = pvm_mytid();
  assert(me >= 0);
  my_parent = pvm_parent();
  if (my_parent == PvmNoParent) {
	/* We are an independent process */
	NumNodes = 1;
	MyNodeNum = 0;
        the_nodelist = (int *)malloc(NumNodes * sizeof(int));
	assert(the_nodelist != (int *)NULL);
	the_nodelist[0] = me;
  }
  else {
	/* Get static configuration from parent */
        retval = pvm_recv(-1, 0);
	assert(retval >= 0);
	pvm_upkint(&NumNodes, 1, 1);
	the_nodelist = (int *)malloc(NumNodes * sizeof(int));
	assert(the_nodelist != (int *)NULL);
	pvm_upkint(the_nodelist, NumNodes, 1);
  }
  for (i = 0; i < NumNodes; i++)
        if (me == the_nodelist[i])
	     break;
  assert(i != NumNodes);
  MyNodeNum = i;
  printf("NumNodes = %d, MyNodeNum = %d\n", NumNodes, MyNodeNum);
#endif
#ifdef IntelNX
  MyNodeNum = mynode();
  NumNodes = numnodes();
#endif
}


void growPorts(int newSize)
{
  int i;
  int oldNumPorts;
  CommPort** oldPorts;

  AcquireSpinlock(thePortsLock);
  if(numPorts < newSize){
	oldNumPorts = numPorts;
	oldPorts = thePorts;
	for(numPorts = oldNumPorts; numPorts<newSize; numPorts*=2)    ;
	thePorts = (CommPort**)malloc(numPorts*sizeof(CommPort*));
	assert(thePorts);
	for(i=0; i<numPorts; ++i){
	  if(i < oldNumPorts){
		thePorts[i] = oldPorts[i];
	  }else{
		thePorts[i] = (CommPort*)malloc(sizeof(CommPort));
		assert(thePorts[i]);
		InitSpinlock(thePorts[i]->lock);
		thePorts[i]->waiters = (Thread*)0;
		thePorts[i]->firstBuf = thePorts[i]->lastBuf = (BufferDescriptor*)0;
	  }
	}
	free(oldPorts);
  }
  ReleaseSpinlock(thePortsLock);
}


void Tsend(void* sendBuf, int numBytes, int destNode, int destPort)
{
  CommPort* port;
  BufferDescriptor* bufDesc;
  Thread* wakee;
#ifdef PVM3
  int retval;
#endif

  if(destNode == MyNodeNum){
	if(destPort > numPorts)    growPorts(destPort);
	AcquireSpinlock(thePortsLock);
	port = thePorts[destPort];
	ReleaseSpinlock(thePortsLock);
	AcquireSpinlock(port->lock);
	if(port->waiters){
	  wakee = port->waiters;
	  port->waiters = wakee->next;
	  if(numBytes > wakee->msgSize){
		numBytes = wakee->msgSize;
	  }else{
		wakee->msgSize = numBytes;
	  }
	  memcpy(wakee->msgBuf, sendBuf, numBytes);
	  threadUnblock(wakee);
	}else{
	  bufDesc = newBufDesc();
	  bufDesc->buf = malloc(numBytes);
	  assert(bufDesc->buf);
	  bufDesc->size = numBytes;
	  memcpy(bufDesc->buf, sendBuf, numBytes);
	  bufDesc->next = (BufferDescriptor*)0;
	  if(port->firstBuf){
		port->lastBuf->next = bufDesc;
		port->lastBuf = bufDesc;
	  }else{
		port->firstBuf = port->lastBuf = bufDesc;
	  }
	}
	ReleaseSpinlock(port->lock);
  }else{
#ifdef IntelNX
	csend(destPort, sendBuf, numBytes, destNode, 0);
#endif
#ifdef PVM
	initsend();
	putbytes(sendBuf, numBytes);
	snd(componentName, destNode, destPort);
#endif
#ifdef PVM3
	pvm_initsend(PvmDataDefault);
	pvm_pkbyte((void *)sendBuf, numBytes, 1);
	retval = pvm_send(the_nodelist[destNode], destPort);
	assert(retval >= 0);
#endif
  }
}


int Trecv(void* recvBuf, int maxSize, int portNum)
{
  int newSize;
  CommPort* port;
  BufferDescriptor* bufDesc;

  if(portNum > numPorts)	growPorts(portNum);
  AcquireSpinlock(thePortsLock);
  port = thePorts[portNum];
  ReleaseSpinlock(thePortsLock);
  AcquireSpinlock(port->lock);
  bufDesc = port->firstBuf;
  if(bufDesc){
	if(bufDesc->size < maxSize)    maxSize = bufDesc->size;
	memcpy(recvBuf, bufDesc->buf, maxSize);
	free(bufDesc->buf);
	port->firstBuf = bufDesc->next;
	if(port->firstBuf == (BufferDescriptor*)0){
	  port->lastBuf = (BufferDescriptor*)0;
	}
	freeBufDesc(bufDesc);
	ReleaseSpinlock(port->lock);
	return maxSize;
  }else{
	thisThread->msgBuf = recvBuf;
	thisThread->msgSize = maxSize;
    thisThread->next = port->waiters;
	port->waiters = thisThread;
	ReleaseSpinlock(port->lock);
	AcquireSpinlock(numRecvWaitersLock);
	++numRecvWaiters;
	ReleaseSpinlock(numRecvWaitersLock);
	threadBlock();
	AcquireSpinlock(numRecvWaitersLock);
	--numRecvWaiters;
	ReleaseSpinlock(numRecvWaitersLock);
	return thisThread->msgSize;
  }
}


void deliverMessages()
{
  int numBytes;
  int destPort;
  Thread* wakee;
  CommPort* port;
  BufferDescriptor* bufDesc;
#ifdef PVM3
  int bufid;
  int the_tid;
  int retval;
#endif

#ifndef OneNode
#ifdef IntelNX
  while(iprobe(-1) > 0){
	destPort = infotype();
	numBytes = infocount();
#endif
#ifdef PVM
  while(probe(-1) >= 0){
	rcvinfo(&numBytes, &destPort, (char*)0, (int*)0);
#endif
#ifdef PVM3
  while((bufid = pvm_probe(-1, -1)) > 0) {
	retval = pvm_bufinfo(bufid, &numBytes, &destPort, &the_tid);
	assert(retval >= 0);
#endif
	if(destPort > numPorts)    growPorts(destPort);
	AcquireSpinlock(thePortsLock);
	port = thePorts[destPort];
	ReleaseSpinlock(thePortsLock);
	if(port->waiters){
	  wakee = port->waiters;
	  port->waiters = wakee->next;
	  if(numBytes > wakee->msgSize){
		numBytes = wakee->msgSize;
	  }else{
		wakee->msgSize = numBytes;
	  }
#ifdef IntelNX
	  crecv(destPort, wakee->msgBuf, numBytes);
#endif
#ifdef PVM
	  rcv(destPort);
	  getbytes(wakee->msgBuf, numBytes);
#endif
#ifdef PVM3
	  retval = pvm_recv(-1, destPort);
	  assert(retval >= 0);
	  pvm_upkbyte(wakee->msgBuf, numBytes, 1);
#endif
	  threadUnblock(wakee);
	}else{
	  bufDesc = newBufDesc();
	  bufDesc->buf = malloc(numBytes);
	  assert(bufDesc->buf);
	  bufDesc->size = numBytes;
#ifdef IntelNX
	  crecv(destPort, bufDesc->buf, numBytes);
#endif
#ifdef PVM
	  rcv(destPort);
	  getbytes(bufDesc->buf, numBytes);
#endif	  
#ifdef PVM3
	  retval = pvm_recv(-1, destPort);
	  assert(retval >= 0);
	  pvm_upkbyte(bufDesc->buf, numBytes, 1);
#endif
	  bufDesc->next = (BufferDescriptor*)0;
	  if(port->firstBuf){
		port->lastBuf->next = bufDesc;
		port->lastBuf = bufDesc;
	  }else{
		port->firstBuf = port->lastBuf = bufDesc;
	  }
	}
  }
#endif /*ndef OneNode*/
}





